/*
 * Copyright (C) 2016 MediaTek Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#define DEBUG 1

#include <linux/debugfs.h>
#include <linux/blkdev.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/cpumask.h>
#include <linux/cputime.h>
#include <linux/tick.h>
#include <linux/kernel_stat.h>
#include <linux/sched.h>
#include <linux/mutex.h>
#include <linux/spinlock.h>
#include <linux/spinlock_types.h>

#include <linux/vmalloc.h>
#include <linux/memblock.h>
#include <linux/blk_types.h>
#include <linux/module.h>

#ifdef CONFIG_MTK_USE_RESERVED_EXT_MEM
#include <linux/exm_driver.h>
#endif

#include <mt-plat/mtk_blocktag.h>

/* debugfs dentries */
struct dentry *mtk_btag_droot;
struct dentry *mtk_btag_dlog;
static void mtk_btag_init_debugfs(void);

/* blocktag */
static DEFINE_MUTEX(mtk_btag_list_lock);
static LIST_HEAD(mtk_btag_list);

static struct mtk_blocktag *mtk_btag_find(const char *name)
{
	struct mtk_blocktag *btag, *n;

	list_for_each_entry_safe(btag, n, &mtk_btag_list, list) {
		if (!strncmp(btag->name, name, BLOCKTAG_NAME_LEN-1))
			return btag;
	}
	return NULL;
}

static struct mtk_blocktag *mtk_btag_find_locked(const char *name)
{
	struct mtk_blocktag *btag;

	mutex_lock(&mtk_btag_list_lock);
	btag = mtk_btag_find(name);
	mutex_unlock(&mtk_btag_list_lock);
	return btag;
}

/* pid logger: page loger*/
unsigned long long mtk_btag_system_dram_size;
struct page_pid_logger *mtk_btag_pagelogger;
spinlock_t mtk_btag_pagelogger_lock;

static size_t mtk_btag_seq_pidlog_usedmem(struct seq_file *seq)
{
	size_t size = 0;

	if (!IS_ERR_OR_NULL(mtk_btag_pagelogger)) {
		size = (sizeof(struct page_pid_logger)*(mtk_btag_system_dram_size >> PAGE_SHIFT));
		seq_printf(seq, "page pid logger buffer: %llu entries * %zu = %zu bytes\n",
			(mtk_btag_system_dram_size >> PAGE_SHIFT),
			sizeof(struct page_pid_logger),
			size);
	}
	return size;
}

#define biolog_fmt "wl:%d%%,%lld,%lld,%d.vm:%lld,%lld,%lld,%lld,%lld.cpu:%llu,%llu,%llu,%llu,%llu,%llu,%llu.pid:%d,"
#define biolog_fmt_wt "wt:%d,%d,%lld."
#define biolog_fmt_rt "rt:%d,%d,%lld."
#define pidlog_fmt "{%05d:%05d:%08d:%05d:%08d}"

void mtk_btag_pidlog_insert(struct mtk_btag_pidlogger *pidlog, pid_t pid, __u32 len, int rw)
{
	int i;
	struct mtk_btag_pidlogger_entry *pe;
	struct mtk_btag_pidlogger_entry_rw *prw;

	for (i = 0; i < BLOCKTAG_PIDLOG_ENTRIES; i++) {
		pe = &pidlog->info[i];
		if ((pe->pid == pid) || (pe->pid == 0)) {
			pe->pid = pid;
			prw = (rw) ? &pe->w : &pe->r;
			prw->count++;
			prw->length += len;
			break;
		}
	}
}
EXPORT_SYMBOL_GPL(mtk_btag_pidlog_insert);

static void mtk_btag_pidlog_add(struct request_queue *q, struct bio *bio,
	unsigned short pid, __u32 len)
{
	int major = bio->bi_bdev ? MAJOR(bio->bi_bdev->bd_dev) : 0;

	if (pid != 0xFFFF && major) {
#ifdef CONFIG_MTK_UFS_BLOCK_IO_LOG
		if (major == SCSI_DISK0_MAJOR || major == BLOCK_EXT_MAJOR) {
			mtk_btag_pidlog_add_ufs(q, pid, len, rw);
			return;
		}
#endif
#ifdef CONFIG_MMC_BLOCK_IO_LOG
		if (major == MMC_BLOCK_MAJOR || major == BLOCK_EXT_MAJOR) {
			mtk_btag_pidlog_add_mmc(q, pid, len, rw);
			return;
		}
#endif
	}
}

/*
 * pidlog: hook function for __blk_bios_map_sg()
 * rw: 0=read, 1=write
 */
void mtk_btag_pidlog_map_sg(struct request_queue *q, struct bio *bio, struct bio_vec *bvec)
{
	struct page_pid_logger *ppl, tmp;
	unsigned long page_offset;
	unsigned long flags;

	if (!mtk_btag_pagelogger || !bio || !bvec)
		return;

	page_offset = (unsigned long)(__page_to_pfn(bvec->bv_page)) - PHYS_PFN_OFFSET;
	spin_lock_irqsave(&mtk_btag_pagelogger_lock, flags);
	ppl = ((struct page_pid_logger *)mtk_btag_pagelogger) + page_offset;
	tmp.pid1 = ppl->pid1;
	tmp.pid2 = ppl->pid2;
	spin_unlock_irqrestore(&mtk_btag_pagelogger_lock, flags);

	mtk_btag_pidlog_add(q, bio, tmp.pid1, bvec->bv_len);
	mtk_btag_pidlog_add(q, bio, tmp.pid2, bvec->bv_len);
}
EXPORT_SYMBOL_GPL(mtk_btag_pidlog_map_sg);

/* pidlog: hook function for submit_bio() */
void mtk_btag_pidlog_submit_bio(struct bio *bio)
{
	struct bio_vec bvec;
	struct bvec_iter iter;

	if (!mtk_btag_pagelogger)
		return;

	bio_for_each_segment(bvec, bio, iter) {
		struct page_pid_logger *ppl;
		unsigned long flags;

		if (bvec.bv_page) {
			unsigned long page_index;

			page_index = (unsigned long)(__page_to_pfn(bvec.bv_page)) - PHYS_PFN_OFFSET;
			ppl = ((struct page_pid_logger *)mtk_btag_pagelogger) + page_index;
			spin_lock_irqsave(&mtk_btag_pagelogger_lock, flags);
			if (page_index < (mtk_btag_system_dram_size >> PAGE_SHIFT)) {
				if (ppl->pid1 == 0XFFFF && ppl->pid2 != current->pid)
					ppl->pid1 = current->pid;
				else if (ppl->pid1 != current->pid)
					ppl->pid2 = current->pid;
			}
			spin_unlock_irqrestore(&mtk_btag_pagelogger_lock, flags);
		}
	}
}
EXPORT_SYMBOL_GPL(mtk_btag_pidlog_submit_bio);

/* pidlog: hook function for filesystem's write_begin() */
void mtk_btag_pidlog_write_begin(struct page *p)
{
	struct page_pid_logger *ppl;
	unsigned long flags;
	unsigned long page_index;

	if (!mtk_btag_pagelogger || !p)
		return;

	page_index = (unsigned long)(__page_to_pfn(p)) - PHYS_PFN_OFFSET;
	ppl = ((struct page_pid_logger *)mtk_btag_pagelogger) + page_index;
	spin_lock_irqsave(&mtk_btag_pagelogger_lock, flags);
	if (page_index < (mtk_btag_system_dram_size >> PAGE_SHIFT)) {
		if (ppl->pid1 == 0XFFFF)
			ppl->pid1 = current->pid;
		else if (ppl->pid1 != current->pid)
			ppl->pid2 = current->pid;
	}
	spin_unlock_irqrestore(&mtk_btag_pagelogger_lock, flags);
}
EXPORT_SYMBOL_GPL(mtk_btag_pidlog_write_begin);

/* evaluate vmstat trace from global_page_state() */
void mtk_btag_vmstat_eval(struct mtk_btag_vmstat *vm)
{
	vm->file_pages = ((global_page_state(NR_FILE_PAGES)) << (PAGE_SHIFT - 10));
	vm->file_dirty = ((global_page_state(NR_FILE_DIRTY)) << (PAGE_SHIFT - 10));
	vm->dirtied = ((global_page_state(NR_DIRTIED))	<< (PAGE_SHIFT - 10));
	vm->writeback = ((global_page_state(NR_WRITEBACK))	<< (PAGE_SHIFT - 10));
	vm->written = ((global_page_state(NR_WRITTEN))	<< (PAGE_SHIFT - 10));
}
EXPORT_SYMBOL_GPL(mtk_btag_vmstat_eval);

/* evaluate pidlog trace from context */
void mtk_btag_pidlog_eval(struct mtk_btag_pidlogger *pl, struct mtk_btag_pidlogger *ctx_pl)
{
	int i;

	for (i = 0; i < BLOCKTAG_PIDLOG_ENTRIES; i++) {
		if (ctx_pl->info[i].pid == 0)
			break;
	}

	if (i != 0) {
		int size = i * sizeof(struct mtk_btag_pidlogger_entry);

		memcpy(&pl->info[0], &ctx_pl->info[0], size);
		memset(&ctx_pl->info[0], 0, size);
	}
}
EXPORT_SYMBOL_GPL(mtk_btag_pidlog_eval);

static __u64 mtk_btag_cpu_idle_time(int cpu)
{
	u64 idle, idle_time = -1ULL;

	if (cpu_online(cpu))
		idle_time = get_cpu_idle_time_us(cpu, NULL);

	if (idle_time == -1ULL)
		/* !NO_HZ or cpu offline so we can rely on cpustat.idle */
		idle = kcpustat_cpu(cpu).cpustat[CPUTIME_IDLE];
	else
		idle = usecs_to_cputime64(idle_time);

	return idle;
}

static __u64 mtk_btag_cpu_iowait_time(int cpu)
{
	__u64 iowait, iowait_time = -1ULL;

	if (cpu_online(cpu))
		iowait_time = get_cpu_iowait_time_us(cpu, NULL);

	if (iowait_time == -1ULL)
		/* !NO_HZ or cpu offline so we can rely on cpustat.iowait */
		iowait = kcpustat_cpu(cpu).cpustat[CPUTIME_IOWAIT];
	else
		iowait = usecs_to_cputime64(iowait_time);

	return iowait;
}

/* evaluate cpu trace from kcpustat_cpu() */
void mtk_btag_cpu_eval(struct mtk_btag_cpu *cpu)
{
	int i;
	__u64 user, nice, system, idle, iowait, irq, softirq;

	user = nice = system = idle = iowait = irq = softirq = 0;

	for_each_possible_cpu(i) {
		user += kcpustat_cpu(i).cpustat[CPUTIME_USER];
		nice += kcpustat_cpu(i).cpustat[CPUTIME_NICE];
		system += kcpustat_cpu(i).cpustat[CPUTIME_SYSTEM];
		idle += mtk_btag_cpu_idle_time(i);
		iowait += mtk_btag_cpu_iowait_time(i);
		irq += kcpustat_cpu(i).cpustat[CPUTIME_IRQ];
		softirq += kcpustat_cpu(i).cpustat[CPUTIME_SOFTIRQ];
	}

	cpu->user = cputime64_to_clock_t(user);
	cpu->nice = cputime64_to_clock_t(nice);
	cpu->system = cputime64_to_clock_t(system);
	cpu->idle = cputime64_to_clock_t(idle);
	cpu->iowait = cputime64_to_clock_t(iowait);
	cpu->irq = cputime64_to_clock_t(irq);
	cpu->softirq = cputime64_to_clock_t(softirq);
}
EXPORT_SYMBOL_GPL(mtk_btag_cpu_eval);

static void mtk_btag_throughput_rw_eval(struct mtk_btag_throughput_rw *rw)
{
	__u64 usage;

	usage = rw->usage;

	do_div(usage, 1000000); /* convert ns to ms */

	if (usage && rw->size) {
		rw->speed = (rw->size) / (__u32)usage;  /* bytes/ms */
		rw->speed = (rw->speed*1000) >> 10; /* KB/s */
		rw->usage = usage;
	} else {
		rw->speed = 0;
		rw->size = 0;
		rw->usage = 0;
	}
}
/* calculate throughput */
void mtk_btag_throughput_eval(struct mtk_btag_throughput *tp)
{
	mtk_btag_throughput_rw_eval(&tp->r);
	mtk_btag_throughput_rw_eval(&tp->w);
}
EXPORT_SYMBOL_GPL(mtk_btag_throughput_eval);

/* print trace to kerne log */
static void mtk_btag_klog_entry(char **ptr, int *len, struct mtk_btag_trace *tr)
{
	int i, n;

#define boundary_check() { *len -= n; *ptr += n; }

	if (tr->throughput.r.usage) {
		n = snprintf(*ptr, *len, biolog_fmt_rt,
			tr->throughput.r.speed,
			tr->throughput.r.size,
			tr->throughput.r.usage);
		boundary_check();
		if (*len < 0)
			return;
	}

	if (tr->throughput.w.usage) {
		n = snprintf(*ptr, *len, biolog_fmt_wt,
			tr->throughput.w.speed,
			tr->throughput.w.size,
			tr->throughput.w.usage);
		boundary_check();
		if (*len < 0)
			return;
	}

	n = snprintf(*ptr, *len, biolog_fmt,
		tr->workload.percent,
		tr->workload.usage,
		tr->workload.period,
		tr->workload.count,
		tr->vmstat.file_pages,
		tr->vmstat.file_dirty,
		tr->vmstat.dirtied,
		tr->vmstat.writeback,
		tr->vmstat.written,
		tr->cpu.user,
		tr->cpu.nice,
		tr->cpu.system,
		tr->cpu.idle,
		tr->cpu.iowait,
		tr->cpu.irq,
		tr->cpu.softirq,
		tr->pid);
	boundary_check();
	if (*len < 0)
		return;

	for (i = 0; i < BLOCKTAG_PIDLOG_ENTRIES; i++) {
		struct mtk_btag_pidlogger_entry *pe;

		pe = &tr->pidlog.info[i];

		if (pe->pid == 0)
			break;

		n = snprintf(*ptr, *len, pidlog_fmt,
			pe->pid,
			pe->w.count,
			pe->w.length,
			pe->r.count,
			pe->r.length);
		boundary_check();
		if (*len < 0)
			return;
	}
}

void mtk_btag_klog(struct mtk_blocktag *btag, struct mtk_btag_trace *tr)
{
	int len;
	char *ptr;
	unsigned long flags;

	if (!btag || !btag->klog_enable || !tr)
		return;

	len = BLOCKTAG_PRINT_LEN-1;
	ptr = &btag->prbuf.buf[0];

	spin_lock_irqsave(&btag->prbuf.lock, flags);
	mtk_btag_klog_entry(&ptr, &len, tr);
	spin_unlock_irqrestore(&btag->prbuf.lock, flags);
}
EXPORT_SYMBOL_GPL(mtk_btag_klog);

static int mtk_btag_pr_time(char *out, int size, const char *str, __u64 t)
{
	uint32_t nsec;
	int ret;

	nsec = do_div(t, 1000000000);
	ret = snprintf(out, size, ",%s=[%lu.%06lu]", str, (unsigned long)t, (unsigned long)nsec/1000);
	return ret;
}

static const char *mtk_btag_pr_speed(char *out, int size, __u64 usage, __u32 bytes)
{
	__u32 speed;

	if (!usage || !bytes)
		return "";

	do_div(usage, 1000); /* convert ns to us */
	speed = 1000 * bytes / (__u32)usage;  /* bytes/ms */
	speed = (speed*1000) >> 10; /* KB/s */

	snprintf(out, size, ",%u KB/s", speed);
	return out;
}

void mtk_btag_task_timetag(char *buf, unsigned len, unsigned stage,
	unsigned max, const char *name[], uint64_t *t, __u32 bytes)
{
	__u64 busy_time = 0;
	int i;
	int ret;

	if (!buf || !len)
		return;

	for (i = 0; i <= stage; i++) {
		ret = mtk_btag_pr_time(buf, len, name[i], t[i]);
		buf += ret;
		len -= ret;
	}

	if (stage == max-1) {
		busy_time = t[stage] - t[0];
		ret = mtk_btag_pr_time(buf, len, "busy", busy_time);
		buf += ret;
		len -= ret;
		mtk_btag_pr_speed(buf, len, busy_time, bytes);
	}
}
EXPORT_SYMBOL_GPL(mtk_btag_task_timetag);


void mtk_btag_seq_time(struct seq_file *seq, uint64_t time)
{
	uint32_t nsec;

	nsec = do_div(time, 1000000000);
	seq_printf(seq, "[%5lu.%06lu]", (unsigned long)time, (unsigned long)nsec/1000);
}
EXPORT_SYMBOL_GPL(mtk_btag_seq_time);

static void mtk_btag_seq_trace(struct seq_file *seq, const char *name, struct mtk_btag_trace *tr)
{
	int i;

	if (tr->time <= 0)
		return;

	mtk_btag_seq_time(seq, tr->time);
	seq_printf(seq, "%s.q:%d.", name, tr->qid);

	if (tr->throughput.r.usage)
		seq_printf(seq, biolog_fmt_rt,
			tr->throughput.r.speed,
			tr->throughput.r.size,
			tr->throughput.r.usage);
	if (tr->throughput.w.usage)
		seq_printf(seq, biolog_fmt_wt,
			tr->throughput.w.speed,
			tr->throughput.w.size,
			tr->throughput.w.usage);

	seq_printf(seq, biolog_fmt,
		tr->workload.percent,
		tr->workload.usage,
		tr->workload.period,
		tr->workload.count,
		tr->vmstat.file_pages,
		tr->vmstat.file_dirty,
		tr->vmstat.dirtied,
		tr->vmstat.writeback,
		tr->vmstat.written,
		tr->cpu.user,
		tr->cpu.nice,
		tr->cpu.system,
		tr->cpu.idle,
		tr->cpu.iowait,
		tr->cpu.irq,
		tr->cpu.softirq,
		tr->pid);

	for (i = 0; i < BLOCKTAG_PIDLOG_ENTRIES; i++) {
		struct mtk_btag_pidlogger_entry *pe;

		pe = &tr->pidlog.info[i];

		if (pe->pid == 0)
			break;

		seq_printf(seq, pidlog_fmt,
			pe->pid,
			pe->w.count,
			pe->w.length,
			pe->r.count,
			pe->r.length);
	}
	seq_puts(seq, ".\n");
}

/* get current trace in debugfs ring buffer */
struct mtk_btag_trace *mtk_btag_curr_trace(struct mtk_btag_ringtrace *rt)
{
	if (rt)
		return &rt->trace[rt->index];
	else
		return NULL;
}
EXPORT_SYMBOL_GPL(mtk_btag_curr_trace);

/* step to next trace in debugfs ring buffer */
struct mtk_btag_trace *mtk_btag_next_trace(struct mtk_btag_ringtrace *rt)
{
	rt->index++;
	if (rt->index >= rt->max)
		rt->index = 0;

	return mtk_btag_curr_trace(rt);
}
EXPORT_SYMBOL_GPL(mtk_btag_next_trace);

/* clear debugfs ring buffer */
static void mtk_btag_clear_trace(struct mtk_btag_ringtrace *rt)
{
	unsigned long flags;

	spin_lock_irqsave(&rt->lock, flags);
	memset(rt->trace, 0, (sizeof(struct mtk_btag_trace) * rt->max));
	rt->index = 0;
	spin_unlock_irqrestore(&rt->lock, flags);
}

static void mtk_btag_seq_debug_show_ringtrace(struct seq_file *seq,
	struct mtk_blocktag *btag)
{
	struct mtk_btag_ringtrace *rt = BTAG_RT(btag);
	unsigned long flags;
	int i, end;

	if (!rt)
		return;

	if (rt->index >= rt->max || rt->index < 0)
		rt->index = 0;

	seq_printf(seq, "<%s: blocktag trace>\n", btag->name);

	spin_lock_irqsave(&rt->lock, flags);
	end = (rt->index > 0) ? rt->index-1 : rt->max-1;
	for (i = rt->index;;) {
		mtk_btag_seq_trace(seq, btag->name, &rt->trace[i]);
		if (i == end)
			break;
		i = (i >= rt->max-1) ? 0 : i+1;
	};
	spin_unlock_irqrestore(&rt->lock, flags);
}


static size_t mtk_btag_seq_sub_show_usedmem(struct seq_file *seq,
	struct mtk_blocktag *btag)
{
	size_t used_mem = 0;
	size_t size;

	seq_printf(seq, "<%s: memory usage>\n", btag->name);
	seq_printf(seq, "%s blocktag: %zu bytes\n", btag->name, sizeof(struct mtk_blocktag));
	used_mem += sizeof(struct mtk_blocktag);

	if (BTAG_RT(btag)) {
		size = (sizeof(struct mtk_btag_trace) * BTAG_RT(btag)->max);
		seq_printf(seq, "%s debug ring buffer: %d traces * %zu = %zu bytes\n",
			btag->name,
			BTAG_RT(btag)->max,
			sizeof(struct mtk_btag_trace),
			size);
		used_mem += size;
	}

	if (BTAG_CTX(btag)) {
		size = btag->ctx.size * btag->ctx.count;
		seq_printf(seq, "%s queue context: %d contexts * %d = %zu bytes\n",
			btag->name,
			btag->ctx.count,
			btag->ctx.size,
			size);
		used_mem += size;
	}
	seq_printf(seq, "%s sub-total: %zu KB\n", btag->name, used_mem >> 10);
	return used_mem;
}

/* clear all ringtraces */
static ssize_t mtk_btag_main_write(struct file *file, const char __user *ubuf, size_t count, loff_t *ppos)
{
	struct mtk_blocktag *btag, *n;

	mutex_lock(&mtk_btag_list_lock);
	list_for_each_entry_safe(btag, n, &mtk_btag_list, list)
		mtk_btag_clear_trace(&btag->rt);
	mutex_unlock(&mtk_btag_list_lock);
	return count;
}

/* clear ringtrace */
static ssize_t mtk_btag_sub_write(struct file *file, const char __user *ubuf, size_t count, loff_t *ppos)
{
	struct seq_file *seq = file->private_data;
	struct mtk_blocktag *btag;

	if (seq && seq->private) {
		btag = seq->private;
		mtk_btag_clear_trace(&btag->rt);
	}
	return count;
}

/* seq file operations */
static void *mtk_btag_seq_debug_start(struct seq_file *seq, loff_t *pos)
{
	unsigned int idx;

	if (*pos < 0 || *pos >= 1)
		return NULL;

	idx = *pos + 1;
	return (void *) ((unsigned long) idx);
}

static void *mtk_btag_seq_debug_next(struct seq_file *seq, void *v, loff_t *pos)
{
	unsigned int idx;

	++*pos;
	if (*pos < 0 || *pos >= 1)
		return NULL;

	idx = *pos + 1;
	return (void *) ((unsigned long) idx);
}

static void mtk_btag_seq_debug_stop(struct seq_file *seq, void *v)
{
}

static int mtk_btag_seq_sub_show(struct seq_file *seq, void *v)
{
	struct mtk_blocktag *btag = seq->private;

	if (btag) {
		mtk_btag_seq_debug_show_ringtrace(seq, btag);
		if (btag->seq_show) {
			seq_printf(seq, "<%s: context info>\n", btag->name);
			btag->seq_show(seq);
		}
		mtk_btag_seq_sub_show_usedmem(seq, btag);
	}
	return 0;
}

static const struct seq_operations mtk_btag_seq_sub_ops = {
	.start  = mtk_btag_seq_debug_start,
	.next   = mtk_btag_seq_debug_next,
	.stop   = mtk_btag_seq_debug_stop,
	.show   = mtk_btag_seq_sub_show,
};

static int mtk_btag_sub_open(struct inode *inode, struct file *file)
{
	int rc;

	rc = seq_open(file, &mtk_btag_seq_sub_ops);

	if (!rc) {
		struct seq_file *m = file->private_data;
		struct dentry *entry = container_of(inode->i_dentry.first, struct dentry, d_u.d_alias);

		if (entry && entry->d_parent) {
			pr_err("[BLOCK_TAG] %s: %s/%s\n", __func__,
				entry->d_parent->d_name.name,
				entry->d_name.name);
			m->private = mtk_btag_find_locked(entry->d_parent->d_name.name);
		}
	}
	return rc;
}

static const struct file_operations mtk_btag_sub_fops = {
	.owner		= THIS_MODULE,
	.open		= mtk_btag_sub_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
	.write		= mtk_btag_sub_write,
};

struct mtk_blocktag *mtk_btag_alloc(const char *name,
	unsigned ringtrace_count, size_t ctx_size, unsigned ctx_count,
	mtk_btag_seq_f seq_show)
{
	struct mtk_blocktag *btag;

	if (!name || !ringtrace_count || !ctx_size || !ctx_count)
		return NULL;

	btag = mtk_btag_find(name);
	if (btag) {
		pr_warn("[BLOCK_TAG] %s: blocktag %s already exists.\n", __func__, name);
		return NULL;
	}

	btag = kmalloc(sizeof(struct mtk_blocktag), GFP_NOFS);
	if (!btag)
		return NULL;

	memset(btag, 0, sizeof(struct mtk_blocktag));
	btag->seq_show = seq_show;
	btag->used_mem = sizeof(struct mtk_blocktag) +
		(sizeof(struct mtk_btag_trace) * ringtrace_count) +
		(ctx_count * ctx_size);

	/* ringtrace */
	btag->rt.index = 0;
	btag->rt.max = ringtrace_count;
	spin_lock_init(&btag->rt.lock);
	btag->rt.trace = kmalloc_array(ringtrace_count,
		sizeof(struct mtk_btag_trace), GFP_NOFS);
	if (!btag->rt.trace) {
		kfree(btag);
		return NULL;
	}
	memset(btag->rt.trace, 0, (sizeof(struct mtk_btag_trace) * ringtrace_count));
	strncpy(btag->name, name, BLOCKTAG_NAME_LEN-1);
	spin_lock_init(&btag->rt.lock);

	/* context */
	btag->ctx.count = ctx_count;
	btag->ctx.size = ctx_size;
	btag->ctx.priv = kmalloc_array(ctx_count, ctx_size, GFP_NOFS);
	if (!btag->ctx.priv) {
		kfree(btag->rt.trace);
		kfree(btag);
		return NULL;
	}
	memset(btag->ctx.priv, 0, ctx_size * ctx_count);

	/* debugfs dentries */
	mtk_btag_init_debugfs();
	btag->dentry.droot = debugfs_create_dir(name, mtk_btag_droot);

	if (IS_ERR(btag->dentry.droot)) {
		pr_warn("[BLOCK_TAG] %s: fail to create debugfs root\n", name);
		goto out;
	}

	btag->dentry.dmem = debugfs_create_u32("used_mem", 0440, btag->dentry.droot, &btag->used_mem);

	if (IS_ERR(btag->dentry.dmem))
		pr_warn("[BLOCK_TAG] %s: fail to create used_mem at debugfs\n", name);

	btag->dentry.dklog = debugfs_create_u32("klog_enable", 0660, btag->dentry.droot, &btag->klog_enable);

	if (IS_ERR(btag->dentry.dklog))
		pr_warn("[BLOCK_TAG] %s: fail to create klog_enable at debugfs\n", name);

	btag->dentry.dlog = debugfs_create_file("blockio", S_IFREG | S_IRUGO,
		btag->dentry.droot, (void *)0, &mtk_btag_sub_fops);

	if (IS_ERR(btag->dentry.dlog))
		pr_warn("[BLOCK_TAG] %s: fail to create blockio at debugfs\n", name);

out:
	spin_lock_init(&btag->prbuf.lock);
	list_add(&btag->list, &mtk_btag_list);

	return btag;
}
EXPORT_SYMBOL_GPL(mtk_btag_alloc);

void mtk_btag_free(struct mtk_blocktag *btag)
{
	if (!btag)
		return;

	list_del(&btag->list);
	debugfs_remove_recursive(btag->dentry.droot);
	kfree(btag->ctx.priv);
	kfree(btag->rt.trace);
	kfree(btag);
}
EXPORT_SYMBOL_GPL(mtk_btag_free);

static int __init mtk_btag_early_memory_info(void)
{
	phys_addr_t start, end;

	start = memblock_start_of_DRAM();
	end = memblock_end_of_DRAM();
	mtk_btag_system_dram_size = (unsigned long long)(end - start);
	pr_debug("[BLOCK_TAG] DRAM: %pa - %pa, size: 0x%llx\n", &start,
		&end, (unsigned long long)(end - start));
	return 0;
}
fs_initcall(mtk_btag_early_memory_info);

static void mtk_btag_pidlogger_init(void)
{
	unsigned long count = mtk_btag_system_dram_size >> PAGE_SHIFT;
	unsigned long size = count * sizeof(struct page_pid_logger);

	if (mtk_btag_pagelogger)
		goto init;

	spin_lock_init(&mtk_btag_pagelogger_lock);

#ifdef CONFIG_MTK_USE_RESERVED_EXT_MEM
	mtk_btag_pagelogger = extmem_malloc_page_align(size);
#else
	mtk_btag_pagelogger = vmalloc(size);
#endif

init:
	if (mtk_btag_pagelogger)
		memset(mtk_btag_pagelogger, -1, size);
	else
		pr_info("[BLOCK_TAG] blockio: fail to allocate mtk_btag_pagelogger\n");
}

static int mtk_btag_seq_main_show(struct seq_file *seq, void *v)
{
	size_t used_mem = 0;
	struct mtk_blocktag *btag, *n;

	seq_puts(seq, "[Trace]\n");
	mutex_lock(&mtk_btag_list_lock);
	list_for_each_entry_safe(btag, n, &mtk_btag_list, list)
		mtk_btag_seq_debug_show_ringtrace(seq, btag);

	seq_puts(seq, "[Info]\n");
	list_for_each_entry_safe(btag, n, &mtk_btag_list, list)
		if (btag->seq_show) {
			seq_printf(seq, "<%s: context info>\n", btag->name);
			btag->seq_show(seq);
		}

	seq_puts(seq, "[Memory Usage]\n");
	list_for_each_entry_safe(btag, n, &mtk_btag_list, list)
		used_mem += mtk_btag_seq_sub_show_usedmem(seq, btag);
	mutex_unlock(&mtk_btag_list_lock);

	seq_puts(seq, "<blocktag core>\n");
	used_mem += mtk_btag_seq_pidlog_usedmem(seq);

	seq_puts(seq, "--------------------------------\n");
	seq_printf(seq, "Total: %zu KB\n", used_mem >> 10);

	return 0;
}

static const struct seq_operations mtk_btag_seq_main_ops = {
	.start  = mtk_btag_seq_debug_start,
	.next   = mtk_btag_seq_debug_next,
	.stop   = mtk_btag_seq_debug_stop,
	.show   = mtk_btag_seq_main_show,
};

static int mtk_btag_main_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &mtk_btag_seq_main_ops);
}

static const struct file_operations mtk_btag_main_fops = {
	.owner		= THIS_MODULE,
	.open		= mtk_btag_main_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
	.write		= mtk_btag_main_write,
};

static void mtk_btag_init_debugfs(void)
{
	if (mtk_btag_droot)
		return;

	mtk_btag_droot = debugfs_create_dir("blocktag", NULL);
	if (IS_ERR(mtk_btag_droot)) {
		pr_warn("[BLOCK_TAG] fail to create debugfs root: blocktag\n");
		mtk_btag_droot = NULL;
		return;
	}

	mtk_btag_dlog = debugfs_create_file("blockio", S_IFREG | S_IRUGO, NULL,
		(void *)0, &mtk_btag_main_fops);
	if (IS_ERR(mtk_btag_dlog))
		pr_warn("[BLOCK_TAG] blocktag: fail to create log at debugfs\n");
}

static int __init mtk_btag_init(void)
{
	mtk_btag_pidlogger_init();
	mtk_btag_init_debugfs();
	return 0;
}

static void __exit mtk_btag_exit(void)
{
	debugfs_remove(mtk_btag_dlog);
}

module_init(mtk_btag_init);
module_exit(mtk_btag_exit);

MODULE_AUTHOR("Perry Hsu <perry.hsu@mediatek.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Storage Block Tag Trace");

